try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Choice, Constants
except Exception:
	from ..exception import SDKException
	from ..util import Choice, Constants


class CallBack(object):
	def __init__(self):
		"""Creates an instance of CallBack"""

		self.__url = None
		self.__method = None
		self.__key_modified = dict()

	def get_url(self):
		"""
		The method to get the url

		Returns:
			string: A string representing the url
		"""

		return self.__url

	def set_url(self, url):
		"""
		The method to set the value to url

		Parameters:
			url (string) : A string representing the url
		"""

		if url is not None and not isinstance(url, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: url EXPECTED TYPE: str', None, None)
		
		self.__url = url
		self.__key_modified['url'] = 1

	def get_method(self):
		"""
		The method to get the method

		Returns:
			Choice: An instance of Choice
		"""

		return self.__method

	def set_method(self, method):
		"""
		The method to set the value to method

		Parameters:
			method (Choice) : An instance of Choice
		"""

		if method is not None and not isinstance(method, Choice):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: method EXPECTED TYPE: Choice', None, None)
		
		self.__method = method
		self.__key_modified['method'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Choice, Constants
except Exception:
	from ..exception import SDKException
	from ..util import Choice, Constants


class Format(object):
	def __init__(self):
		"""Creates an instance of Format"""

		self.__decimal_separator = None
		self.__thousand_separator = None
		self.__decimal_places = None
		self.__key_modified = dict()

	def get_decimal_separator(self):
		"""
		The method to get the decimal_separator

		Returns:
			Choice: An instance of Choice
		"""

		return self.__decimal_separator

	def set_decimal_separator(self, decimal_separator):
		"""
		The method to set the value to decimal_separator

		Parameters:
			decimal_separator (Choice) : An instance of Choice
		"""

		if decimal_separator is not None and not isinstance(decimal_separator, Choice):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: decimal_separator EXPECTED TYPE: Choice', None, None)
		
		self.__decimal_separator = decimal_separator
		self.__key_modified['decimal_separator'] = 1

	def get_thousand_separator(self):
		"""
		The method to get the thousand_separator

		Returns:
			Choice: An instance of Choice
		"""

		return self.__thousand_separator

	def set_thousand_separator(self, thousand_separator):
		"""
		The method to set the value to thousand_separator

		Parameters:
			thousand_separator (Choice) : An instance of Choice
		"""

		if thousand_separator is not None and not isinstance(thousand_separator, Choice):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: thousand_separator EXPECTED TYPE: Choice', None, None)
		
		self.__thousand_separator = thousand_separator
		self.__key_modified['thousand_separator'] = 1

	def get_decimal_places(self):
		"""
		The method to get the decimal_places

		Returns:
			Choice: An instance of Choice
		"""

		return self.__decimal_places

	def set_decimal_places(self, decimal_places):
		"""
		The method to set the value to decimal_places

		Parameters:
			decimal_places (Choice) : An instance of Choice
		"""

		if decimal_places is not None and not isinstance(decimal_places, Choice):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: decimal_places EXPECTED TYPE: Choice', None, None)
		
		self.__decimal_places = decimal_places
		self.__key_modified['decimal_places'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
    import importlib
    import logging
    import re
    import json
    from zcrmsdk.src.com.zoho.crm.api.util import Converter, Constants, StreamWrapper
    from zcrmsdk.src.com.zoho.crm.api.exception.sdk_exception import SDKException

except Exception:
    import importlib
    import logging
    import re
    from .converter import Converter
    from .constants import Constants
    from .stream_wrapper import StreamWrapper
    from ..exception import SDKException


class FormDataConverter(Converter):
    """
    This class to process the upload file and stream.
    """

    logger = logging.getLogger('SDKLogger')

    def __init__(self, common_api_handler):

        super().__init__(common_api_handler)
        self.unique_dict = {}
        self.common_api_handler = common_api_handler

    def form_request(self, request_instance, pack, instance_number, class_member_detail):
        path_split = str(pack).rpartition(".")
        class_name = self.module_to_class(path_split[-1])
        pack = path_split[0] + "." + class_name

        try:
            from zcrmsdk.src.com.zoho.crm.api.initializer import Initializer
        except Exception:
            from ..initializer import Initializer

        class_detail = dict(Initializer.json_details[str(pack)])
        request = dict()

        for member_name, member_detail in class_detail.items():
            modification = None

            if (Constants.READ_ONLY in member_detail and bool(
                    member_detail[Constants.READ_ONLY])) or Constants.NAME not in member_detail:
                continue

            try:
                modification = getattr(request_instance, Constants.IS_KEY_MODIFIED)(member_name)
            except Exception as e:
                raise SDKException(code=Constants.EXCEPTION_IS_KEY_MODIFIED, cause=e)

            if (modification is None or modification == 0) and Constants.REQUIRED in member_detail and bool(
                    member_detail[Constants.REQUIRED]):
                raise SDKException(Constants.MANDATORY_VALUE_ERROR,
                                   Constants.MANDATORY_KEY_ERROR + member_name)

            field_value = getattr(request_instance, self.construct_private_member(class_name=class_name, member_name=member_name))

            if modification is not None and modification != 0 and self.value_checker(
                    class_name=class_name,
                    member_name=member_name,
                    key_details=member_detail,
                    value=field_value,
                    unique_values_map=self.unique_dict,
                    instance_number=instance_number) is True:
                key_name = member_detail.get(Constants.NAME)
                data_type = member_detail.get(Constants.TYPE)

                if data_type == Constants.LIST_NAMESPACE:
                    request[key_name] = self.set_json_array(field_value, member_detail)

                elif data_type == Constants.MAP_NAMESPACE:
                    request[key_name] = self.set_json_object(field_value, member_detail)

                elif Constants.STRUCTURE_NAME in member_detail:
                    request[key_name] = self.form_request(field_value, member_detail.get(Constants.STRUCTURE_NAME), 0, member_detail)

                else:
                    request[key_name] = field_value

            return request

    def append_to_request(self, request_base, request_object):
        form_data_request_body = []
        self.add_file_body(request_object, form_data_request_body)
        request_base.file = True
        return form_data_request_body

    def add_file_body(self, request_object, request_body):
        for key_name, key_value in request_object.items():
            if isinstance(key_value, list):
                for each_object in key_value:
                    if isinstance(each_object, StreamWrapper):
                        request_body.append((key_name, each_object.get_stream()))
                    else:
                        request_body.append((key_name, key_value))
            elif isinstance(key_value, StreamWrapper):
                entry = (key_name, key_value.get_stream())
                request_body.append(entry)
            else:
                entry = (key_name, key_value)
                request_body.append(entry)

    def set_json_object(self, field_value, member_detail):
        json_object = {}
        request_object = dict(field_value)

        if member_detail is None:
            for key, value in request_object.items():
                json_object[key] = self.redirector_for_object_to_json(value)

        else:
            keys_detail = member_detail[Constants.KEYS]

            for key_detail in keys_detail:
                key_name = key_detail[Constants.NAME]
                key_value = None

                if key_name in request_object and request_object[key_name] is not None:
                    if Constants.STRUCTURE_NAME in key_detail:
                        key_value = self.form_request(field_value[key_name], key_detail[Constants.STRUCTURE_NAME], 0, member_detail)

                    else:
                        key_value = self.redirector_for_object_to_json(field_value[key_name])

                    json_object[key_name] = key_value

        return json_object

    def set_json_array(self, field_value, member_detail):

        json_array = []
        request_objects = list(field_value)

        if member_detail is None:
            for request in request_objects:
                json_array.append(self.redirector_for_object_to_json(request))

        else:
            if Constants.STRUCTURE_NAME in member_detail:
                instance_count = 0
                pack = member_detail[Constants.STRUCTURE_NAME]

                for request in request_objects:
                    json_array.append(self.form_request(request, pack, instance_count, member_detail))
                    instance_count += 1

            else:
                for request in field_value:
                    json_array.append(self.redirector_for_object_to_json(request))

        return json_array

    def redirector_for_object_to_json(self, request):
        if isinstance(request, list):
            return self.set_json_array(request, None)

        if isinstance(request, dict):
            return self.set_json_object(request, None)

        else:
            return request

    def get_wrapped_response(self, response, pack):
        return None

    def get_response(self, response, pack):
        return None

    def construct_private_member(self, class_name, member_name):
        return '_' + class_name + '__' + member_name

    def module_to_class(self, module_name):
        class_name = module_name

        if "_" in module_name:
            class_name = ''
            module_split = str(module_name).split('_')
            for each_name in module_split:
                each_name = each_name.capitalize()
                class_name += each_name

        return class_name



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class Formula(object):
	def __init__(self):
		"""Creates an instance of Formula"""

		self.__return_type = None
		self.__expression = None
		self.__key_modified = dict()

	def get_return_type(self):
		"""
		The method to get the return_type

		Returns:
			string: A string representing the return_type
		"""

		return self.__return_type

	def set_return_type(self, return_type):
		"""
		The method to set the value to return_type

		Parameters:
			return_type (string) : A string representing the return_type
		"""

		if return_type is not None and not isinstance(return_type, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: return_type EXPECTED TYPE: str', None, None)
		
		self.__return_type = return_type
		self.__key_modified['return_type'] = 1

	def get_expression(self):
		"""
		The method to get the expression

		Returns:
			string: A string representing the expression
		"""

		return self.__expression

	def set_expression(self, expression):
		"""
		The method to set the value to expression

		Parameters:
			expression (string) : A string representing the expression
		"""

		if expression is not None and not isinstance(expression, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: expression EXPECTED TYPE: str', None, None)
		
		self.__expression = expression
		self.__key_modified['expression'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
	from zcrmsdk.src.com.zoho.crm.api.record import Record
except Exception:
	from ..exception import SDKException
	from ..util import Constants
	from ..record import Record


class User(Record):
	def __init__(self):
		"""Creates an instance of User"""
		super().__init__()


	def get_country(self):
		"""
		The method to get the country

		Returns:
			string: A string representing the country
		"""

		return self.get_key_value('country')

	def set_country(self, country):
		"""
		The method to set the value to country

		Parameters:
			country (string) : A string representing the country
		"""

		if country is not None and not isinstance(country, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: country EXPECTED TYPE: str', None, None)
		
		self.add_key_value('country', country)

	def get_customize_info(self):
		"""
		The method to get the customize_info

		Returns:
			CustomizeInfo: An instance of CustomizeInfo
		"""

		return self.get_key_value('customize_info')

	def set_customize_info(self, customize_info):
		"""
		The method to set the value to customize_info

		Parameters:
			customize_info (CustomizeInfo) : An instance of CustomizeInfo
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.users.customize_info import CustomizeInfo
		except Exception:
			from .customize_info import CustomizeInfo

		if customize_info is not None and not isinstance(customize_info, CustomizeInfo):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: customize_info EXPECTED TYPE: CustomizeInfo', None, None)
		
		self.add_key_value('customize_info', customize_info)

	def get_role(self):
		"""
		The method to get the role

		Returns:
			Role: An instance of Role
		"""

		return self.get_key_value('role')

	def set_role(self, role):
		"""
		The method to set the value to role

		Parameters:
			role (Role) : An instance of Role
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.roles import Role
		except Exception:
			from ..roles import Role

		if role is not None and not isinstance(role, Role):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: role EXPECTED TYPE: Role', None, None)
		
		self.add_key_value('role', role)

	def get_signature(self):
		"""
		The method to get the signature

		Returns:
			string: A string representing the signature
		"""

		return self.get_key_value('signature')

	def set_signature(self, signature):
		"""
		The method to set the value to signature

		Parameters:
			signature (string) : A string representing the signature
		"""

		if signature is not None and not isinstance(signature, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: signature EXPECTED TYPE: str', None, None)
		
		self.add_key_value('signature', signature)

	def get_city(self):
		"""
		The method to get the city

		Returns:
			string: A string representing the city
		"""

		return self.get_key_value('city')

	def set_city(self, city):
		"""
		The method to set the value to city

		Parameters:
			city (string) : A string representing the city
		"""

		if city is not None and not isinstance(city, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: city EXPECTED TYPE: str', None, None)
		
		self.add_key_value('city', city)

	def get_name_format(self):
		"""
		The method to get the name_format

		Returns:
			string: A string representing the name_format
		"""

		return self.get_key_value('name_format')

	def set_name_format(self, name_format):
		"""
		The method to set the value to name_format

		Parameters:
			name_format (string) : A string representing the name_format
		"""

		if name_format is not None and not isinstance(name_format, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: name_format EXPECTED TYPE: str', None, None)
		
		self.add_key_value('name_format', name_format)

	def get_personal_account(self):
		"""
		The method to get the personal_account

		Returns:
			bool: A bool representing the personal_account
		"""

		return self.get_key_value('personal_account')

	def set_personal_account(self, personal_account):
		"""
		The method to set the value to personal_account

		Parameters:
			personal_account (bool) : A bool representing the personal_account
		"""

		if personal_account is not None and not isinstance(personal_account, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: personal_account EXPECTED TYPE: bool', None, None)
		
		self.add_key_value('personal_account', personal_account)

	def get_default_tab_group(self):
		"""
		The method to get the default_tab_group

		Returns:
			string: A string representing the default_tab_group
		"""

		return self.get_key_value('default_tab_group')

	def set_default_tab_group(self, default_tab_group):
		"""
		The method to set the value to default_tab_group

		Parameters:
			default_tab_group (string) : A string representing the default_tab_group
		"""

		if default_tab_group is not None and not isinstance(default_tab_group, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: default_tab_group EXPECTED TYPE: str', None, None)
		
		self.add_key_value('default_tab_group', default_tab_group)

	def get_language(self):
		"""
		The method to get the language

		Returns:
			string: A string representing the language
		"""

		return self.get_key_value('language')

	def set_language(self, language):
		"""
		The method to set the value to language

		Parameters:
			language (string) : A string representing the language
		"""

		if language is not None and not isinstance(language, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: language EXPECTED TYPE: str', None, None)
		
		self.add_key_value('language', language)

	def get_locale(self):
		"""
		The method to get the locale

		Returns:
			string: A string representing the locale
		"""

		return self.get_key_value('locale')

	def set_locale(self, locale):
		"""
		The method to set the value to locale

		Parameters:
			locale (string) : A string representing the locale
		"""

		if locale is not None and not isinstance(locale, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: locale EXPECTED TYPE: str', None, None)
		
		self.add_key_value('locale', locale)

	def get_microsoft(self):
		"""
		The method to get the microsoft

		Returns:
			bool: A bool representing the microsoft
		"""

		return self.get_key_value('microsoft')

	def set_microsoft(self, microsoft):
		"""
		The method to set the value to microsoft

		Parameters:
			microsoft (bool) : A bool representing the microsoft
		"""

		if microsoft is not None and not isinstance(microsoft, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: microsoft EXPECTED TYPE: bool', None, None)
		
		self.add_key_value('microsoft', microsoft)

	def get_isonline(self):
		"""
		The method to get the isonline

		Returns:
			bool: A bool representing the isonline
		"""

		return self.get_key_value('Isonline')

	def set_isonline(self, isonline):
		"""
		The method to set the value to isonline

		Parameters:
			isonline (bool) : A bool representing the isonline
		"""

		if isonline is not None and not isinstance(isonline, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: isonline EXPECTED TYPE: bool', None, None)
		
		self.add_key_value('Isonline', isonline)

	def get_street(self):
		"""
		The method to get the street

		Returns:
			string: A string representing the street
		"""

		return self.get_key_value('street')

	def set_street(self, street):
		"""
		The method to set the value to street

		Parameters:
			street (string) : A string representing the street
		"""

		if street is not None and not isinstance(street, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: street EXPECTED TYPE: str', None, None)
		
		self.add_key_value('street', street)

	def get_currency(self):
		"""
		The method to get the currency

		Returns:
			string: A string representing the currency
		"""

		return self.get_key_value('Currency')

	def set_currency(self, currency):
		"""
		The method to set the value to currency

		Parameters:
			currency (string) : A string representing the currency
		"""

		if currency is not None and not isinstance(currency, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: currency EXPECTED TYPE: str', None, None)
		
		self.add_key_value('Currency', currency)

	def get_alias(self):
		"""
		The method to get the alias

		Returns:
			string: A string representing the alias
		"""

		return self.get_key_value('alias')

	def set_alias(self, alias):
		"""
		The method to set the value to alias

		Parameters:
			alias (string) : A string representing the alias
		"""

		if alias is not None and not isinstance(alias, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: alias EXPECTED TYPE: str', None, None)
		
		self.add_key_value('alias', alias)

	def get_theme(self):
		"""
		The method to get the theme

		Returns:
			Theme: An instance of Theme
		"""

		return self.get_key_value('theme')

	def set_theme(self, theme):
		"""
		The method to set the value to theme

		Parameters:
			theme (Theme) : An instance of Theme
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.users.theme import Theme
		except Exception:
			from .theme import Theme

		if theme is not None and not isinstance(theme, Theme):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: theme EXPECTED TYPE: Theme', None, None)
		
		self.add_key_value('theme', theme)

	def get_state(self):
		"""
		The method to get the state

		Returns:
			string: A string representing the state
		"""

		return self.get_key_value('state')

	def set_state(self, state):
		"""
		The method to set the value to state

		Parameters:
			state (string) : A string representing the state
		"""

		if state is not None and not isinstance(state, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: state EXPECTED TYPE: str', None, None)
		
		self.add_key_value('state', state)

	def get_fax(self):
		"""
		The method to get the fax

		Returns:
			string: A string representing the fax
		"""

		return self.get_key_value('fax')

	def set_fax(self, fax):
		"""
		The method to set the value to fax

		Parameters:
			fax (string) : A string representing the fax
		"""

		if fax is not None and not isinstance(fax, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: fax EXPECTED TYPE: str', None, None)
		
		self.add_key_value('fax', fax)

	def get_country_locale(self):
		"""
		The method to get the country_locale

		Returns:
			string: A string representing the country_locale
		"""

		return self.get_key_value('country_locale')

	def set_country_locale(self, country_locale):
		"""
		The method to set the value to country_locale

		Parameters:
			country_locale (string) : A string representing the country_locale
		"""

		if country_locale is not None and not isinstance(country_locale, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: country_locale EXPECTED TYPE: str', None, None)
		
		self.add_key_value('country_locale', country_locale)

	def get_first_name(self):
		"""
		The method to get the first_name

		Returns:
			string: A string representing the first_name
		"""

		return self.get_key_value('first_name')

	def set_first_name(self, first_name):
		"""
		The method to set the value to first_name

		Parameters:
			first_name (string) : A string representing the first_name
		"""

		if first_name is not None and not isinstance(first_name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: first_name EXPECTED TYPE: str', None, None)
		
		self.add_key_value('first_name', first_name)

	def get_email(self):
		"""
		The method to get the email

		Returns:
			string: A string representing the email
		"""

		return self.get_key_value('email')

	def set_email(self, email):
		"""
		The method to set the value to email

		Parameters:
			email (string) : A string representing the email
		"""

		if email is not None and not isinstance(email, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: email EXPECTED TYPE: str', None, None)
		
		self.add_key_value('email', email)

	def get_reporting_to(self):
		"""
		The method to get the reporting_to

		Returns:
			User: An instance of User
		"""

		return self.get_key_value('Reporting_To')

	def set_reporting_to(self, reporting_to):
		"""
		The method to set the value to reporting_to

		Parameters:
			reporting_to (User) : An instance of User
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.users.user import User
		except Exception:
			from .user import User

		if reporting_to is not None and not isinstance(reporting_to, User):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: reporting_to EXPECTED TYPE: User', None, None)
		
		self.add_key_value('Reporting_To', reporting_to)

	def get_decimal_separator(self):
		"""
		The method to get the decimal_separator

		Returns:
			string: A string representing the decimal_separator
		"""

		return self.get_key_value('decimal_separator')

	def set_decimal_separator(self, decimal_separator):
		"""
		The method to set the value to decimal_separator

		Parameters:
			decimal_separator (string) : A string representing the decimal_separator
		"""

		if decimal_separator is not None and not isinstance(decimal_separator, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: decimal_separator EXPECTED TYPE: str', None, None)
		
		self.add_key_value('decimal_separator', decimal_separator)

	def get_zip(self):
		"""
		The method to get the zip

		Returns:
			string: A string representing the zip
		"""

		return self.get_key_value('zip')

	def set_zip(self, zip):
		"""
		The method to set the value to zip

		Parameters:
			zip (string) : A string representing the zip
		"""

		if zip is not None and not isinstance(zip, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: zip EXPECTED TYPE: str', None, None)
		
		self.add_key_value('zip', zip)

	def get_website(self):
		"""
		The method to get the website

		Returns:
			string: A string representing the website
		"""

		return self.get_key_value('website')

	def set_website(self, website):
		"""
		The method to set the value to website

		Parameters:
			website (string) : A string representing the website
		"""

		if website is not None and not isinstance(website, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: website EXPECTED TYPE: str', None, None)
		
		self.add_key_value('website', website)

	def get_time_format(self):
		"""
		The method to get the time_format

		Returns:
			string: A string representing the time_format
		"""

		return self.get_key_value('time_format')

	def set_time_format(self, time_format):
		"""
		The method to set the value to time_format

		Parameters:
			time_format (string) : A string representing the time_format
		"""

		if time_format is not None and not isinstance(time_format, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: time_format EXPECTED TYPE: str', None, None)
		
		self.add_key_value('time_format', time_format)

	def get_offset(self):
		"""
		The method to get the offset

		Returns:
			int: An int representing the offset
		"""

		return self.get_key_value('offset')

	def set_offset(self, offset):
		"""
		The method to set the value to offset

		Parameters:
			offset (int) : An int representing the offset
		"""

		if offset is not None and not isinstance(offset, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: offset EXPECTED TYPE: int', None, None)
		
		self.add_key_value('offset', offset)

	def get_profile(self):
		"""
		The method to get the profile

		Returns:
			Profile: An instance of Profile
		"""

		return self.get_key_value('profile')

	def set_profile(self, profile):
		"""
		The method to set the value to profile

		Parameters:
			profile (Profile) : An instance of Profile
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.profiles import Profile
		except Exception:
			from ..profiles import Profile

		if profile is not None and not isinstance(profile, Profile):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: profile EXPECTED TYPE: Profile', None, None)
		
		self.add_key_value('profile', profile)

	def get_mobile(self):
		"""
		The method to get the mobile

		Returns:
			string: A string representing the mobile
		"""

		return self.get_key_value('mobile')

	def set_mobile(self, mobile):
		"""
		The method to set the value to mobile

		Parameters:
			mobile (string) : A string representing the mobile
		"""

		if mobile is not None and not isinstance(mobile, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: mobile EXPECTED TYPE: str', None, None)
		
		self.add_key_value('mobile', mobile)

	def get_last_name(self):
		"""
		The method to get the last_name

		Returns:
			string: A string representing the last_name
		"""

		return self.get_key_value('last_name')

	def set_last_name(self, last_name):
		"""
		The method to set the value to last_name

		Parameters:
			last_name (string) : A string representing the last_name
		"""

		if last_name is not None and not isinstance(last_name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: last_name EXPECTED TYPE: str', None, None)
		
		self.add_key_value('last_name', last_name)

	def get_time_zone(self):
		"""
		The method to get the time_zone

		Returns:
			string: A string representing the time_zone
		"""

		return self.get_key_value('time_zone')

	def set_time_zone(self, time_zone):
		"""
		The method to set the value to time_zone

		Parameters:
			time_zone (string) : A string representing the time_zone
		"""

		if time_zone is not None and not isinstance(time_zone, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: time_zone EXPECTED TYPE: str', None, None)
		
		self.add_key_value('time_zone', time_zone)

	def get_zuid(self):
		"""
		The method to get the zuid

		Returns:
			string: A string representing the zuid
		"""

		return self.get_key_value('zuid')

	def set_zuid(self, zuid):
		"""
		The method to set the value to zuid

		Parameters:
			zuid (string) : A string representing the zuid
		"""

		if zuid is not None and not isinstance(zuid, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: zuid EXPECTED TYPE: str', None, None)
		
		self.add_key_value('zuid', zuid)

	def get_confirm(self):
		"""
		The method to get the confirm

		Returns:
			bool: A bool representing the confirm
		"""

		return self.get_key_value('confirm')

	def set_confirm(self, confirm):
		"""
		The method to set the value to confirm

		Parameters:
			confirm (bool) : A bool representing the confirm
		"""

		if confirm is not None and not isinstance(confirm, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: confirm EXPECTED TYPE: bool', None, None)
		
		self.add_key_value('confirm', confirm)

	def get_full_name(self):
		"""
		The method to get the full_name

		Returns:
			string: A string representing the full_name
		"""

		return self.get_key_value('full_name')

	def set_full_name(self, full_name):
		"""
		The method to set the value to full_name

		Parameters:
			full_name (string) : A string representing the full_name
		"""

		if full_name is not None and not isinstance(full_name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: full_name EXPECTED TYPE: str', None, None)
		
		self.add_key_value('full_name', full_name)

	def get_territories(self):
		"""
		The method to get the territories

		Returns:
			list: An instance of list
		"""

		return self.get_key_value('territories')

	def set_territories(self, territories):
		"""
		The method to set the value to territories

		Parameters:
			territories (list) : An instance of list
		"""

		if territories is not None and not isinstance(territories, list):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: territories EXPECTED TYPE: list', None, None)
		
		self.add_key_value('territories', territories)

	def get_phone(self):
		"""
		The method to get the phone

		Returns:
			string: A string representing the phone
		"""

		return self.get_key_value('phone')

	def set_phone(self, phone):
		"""
		The method to set the value to phone

		Parameters:
			phone (string) : A string representing the phone
		"""

		if phone is not None and not isinstance(phone, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: phone EXPECTED TYPE: str', None, None)
		
		self.add_key_value('phone', phone)

	def get_dob(self):
		"""
		The method to get the dob

		Returns:
			string: A string representing the dob
		"""

		return self.get_key_value('dob')

	def set_dob(self, dob):
		"""
		The method to set the value to dob

		Parameters:
			dob (string) : A string representing the dob
		"""

		if dob is not None and not isinstance(dob, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: dob EXPECTED TYPE: str', None, None)
		
		self.add_key_value('dob', dob)

	def get_date_format(self):
		"""
		The method to get the date_format

		Returns:
			string: A string representing the date_format
		"""

		return self.get_key_value('date_format')

	def set_date_format(self, date_format):
		"""
		The method to set the value to date_format

		Parameters:
			date_format (string) : A string representing the date_format
		"""

		if date_format is not None and not isinstance(date_format, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: date_format EXPECTED TYPE: str', None, None)
		
		self.add_key_value('date_format', date_format)

	def get_status(self):
		"""
		The method to get the status

		Returns:
			string: A string representing the status
		"""

		return self.get_key_value('status')

	def set_status(self, status):
		"""
		The method to set the value to status

		Parameters:
			status (string) : A string representing the status
		"""

		if status is not None and not isinstance(status, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: status EXPECTED TYPE: str', None, None)
		
		self.add_key_value('status', status)

	def get_name(self):
		"""
		The method to get the name

		Returns:
			string: A string representing the name
		"""

		return self.get_key_value('name')

	def set_name(self, name):
		"""
		The method to set the value to name

		Parameters:
			name (string) : A string representing the name
		"""

		if name is not None and not isinstance(name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: name EXPECTED TYPE: str', None, None)
		
		self.add_key_value('name', name)



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.parameter_map import ParameterMap
	from zcrmsdk.src.com.zoho.crm.api.util import APIResponse, CommonAPIHandler, Constants
	from zcrmsdk.src.com.zoho.crm.api.param import Param
except Exception:
	from ..exception import SDKException
	from ..parameter_map import ParameterMap
	from ..util import APIResponse, CommonAPIHandler, Constants
	from ..param import Param


class AttachmentsOperations(object):
	def __init__(self, module_api_name, record_id):
		"""
		Creates an instance of AttachmentsOperations with the given parameters

		Parameters:
			module_api_name (string) : A string representing the module_api_name
			record_id (int) : An int representing the record_id
		"""

		if not isinstance(module_api_name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: module_api_name EXPECTED TYPE: str', None, None)
		
		if not isinstance(record_id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: record_id EXPECTED TYPE: int', None, None)
		
		self.__module_api_name = module_api_name
		self.__record_id = record_id


	def download_attachment(self, id):
		"""
		The method to download attachment

		Parameters:
			id (int) : An int representing the id

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		if not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2/'
		api_path = api_path + str(self.__module_api_name)
		api_path = api_path + '/'
		api_path = api_path + str(self.__record_id)
		api_path = api_path + '/Attachments/'
		api_path = api_path + str(id)
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_GET)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_READ)
		try:
			from zcrmsdk.src.com.zoho.crm.api.attachments.response_handler import ResponseHandler
		except Exception:
			from .response_handler import ResponseHandler
		return handler_instance.api_call(ResponseHandler.__module__, 'application/x-download')

	def delete_attachment(self, id):
		"""
		The method to delete attachment

		Parameters:
			id (int) : An int representing the id

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		if not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2/'
		api_path = api_path + str(self.__module_api_name)
		api_path = api_path + '/'
		api_path = api_path + str(self.__record_id)
		api_path = api_path + '/Attachments/'
		api_path = api_path + str(id)
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_DELETE)
		handler_instance.set_category_method(Constants.REQUEST_METHOD_DELETE)
		try:
			from zcrmsdk.src.com.zoho.crm.api.attachments.action_handler import ActionHandler
		except Exception:
			from .action_handler import ActionHandler
		return handler_instance.api_call(ActionHandler.__module__, 'application/json')

	def get_attachments(self, param_instance=None):
		"""
		The method to get attachments

		Parameters:
			param_instance (ParameterMap) : An instance of ParameterMap

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		if param_instance is not None and not isinstance(param_instance, ParameterMap):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: param_instance EXPECTED TYPE: ParameterMap', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2/'
		api_path = api_path + str(self.__module_api_name)
		api_path = api_path + '/'
		api_path = api_path + str(self.__record_id)
		api_path = api_path + '/Attachments'
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_GET)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_READ)
		handler_instance.set_param(param_instance)
		try:
			from zcrmsdk.src.com.zoho.crm.api.attachments.response_handler import ResponseHandler
		except Exception:
			from .response_handler import ResponseHandler
		return handler_instance.api_call(ResponseHandler.__module__, 'application/json')

	def upload_attachment(self, request):
		"""
		The method to upload attachment

		Parameters:
			request (FileBodyWrapper) : An instance of FileBodyWrapper

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.attachments.file_body_wrapper import FileBodyWrapper
		except Exception:
			from .file_body_wrapper import FileBodyWrapper

		if request is not None and not isinstance(request, FileBodyWrapper):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: request EXPECTED TYPE: FileBodyWrapper', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2/'
		api_path = api_path + str(self.__module_api_name)
		api_path = api_path + '/'
		api_path = api_path + str(self.__record_id)
		api_path = api_path + '/Attachments'
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_POST)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_CREATE)
		handler_instance.set_content_type('multipart/form-data')
		handler_instance.set_request(request)
		handler_instance.set_mandatory_checker(True)
		try:
			from zcrmsdk.src.com.zoho.crm.api.attachments.action_handler import ActionHandler
		except Exception:
			from .action_handler import ActionHandler
		return handler_instance.api_call(ActionHandler.__module__, 'application/json')

	def upload_link_attachment(self, param_instance=None):
		"""
		The method to upload link attachment

		Parameters:
			param_instance (ParameterMap) : An instance of ParameterMap

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		if param_instance is not None and not isinstance(param_instance, ParameterMap):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: param_instance EXPECTED TYPE: ParameterMap', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2/'
		api_path = api_path + str(self.__module_api_name)
		api_path = api_path + '/'
		api_path = api_path + str(self.__record_id)
		api_path = api_path + '/Attachments'
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_POST)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_CREATE)
		handler_instance.set_mandatory_checker(True)
		handler_instance.set_param(param_instance)
		try:
			from zcrmsdk.src.com.zoho.crm.api.attachments.action_handler import ActionHandler
		except Exception:
			from .action_handler import ActionHandler
		return handler_instance.api_call(ActionHandler.__module__, 'application/json')

	def delete_attachments(self, param_instance=None):
		"""
		The method to delete attachments

		Parameters:
			param_instance (ParameterMap) : An instance of ParameterMap

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		if param_instance is not None and not isinstance(param_instance, ParameterMap):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: param_instance EXPECTED TYPE: ParameterMap', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2/'
		api_path = api_path + str(self.__module_api_name)
		api_path = api_path + '/'
		api_path = api_path + str(self.__record_id)
		api_path = api_path + '/Attachments'
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_DELETE)
		handler_instance.set_category_method(Constants.REQUEST_METHOD_DELETE)
		handler_instance.set_param(param_instance)
		try:
			from zcrmsdk.src.com.zoho.crm.api.attachments.action_handler import ActionHandler
		except Exception:
			from .action_handler import ActionHandler
		return handler_instance.api_call(ActionHandler.__module__, 'application/json')


class GetAttachmentsParam(object):
	fields = Param('fields', 'com.zoho.crm.api.Attachments.GetAttachmentsParam')
	page = Param('page', 'com.zoho.crm.api.Attachments.GetAttachmentsParam')
	per_page = Param('per_page', 'com.zoho.crm.api.Attachments.GetAttachmentsParam')


class UploadLinkAttachmentParam(object):
	attachmenturl = Param('attachmentUrl', 'com.zoho.crm.api.Attachments.UploadLinkAttachmentParam')


class DeleteAttachmentsParam(object):
	ids = Param('ids', 'com.zoho.crm.api.Attachments.DeleteAttachmentsParam')



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import StreamWrapper, Constants
	from zcrmsdk.src.com.zoho.crm.api.record.response_handler import ResponseHandler
	from zcrmsdk.src.com.zoho.crm.api.record.download_handler import DownloadHandler
except Exception:
	from ..exception import SDKException
	from ..util import StreamWrapper, Constants
	from .response_handler import ResponseHandler
	from .download_handler import DownloadHandler


class FileBodyWrapper(ResponseHandler, DownloadHandler):
	def __init__(self):
		"""Creates an instance of FileBodyWrapper"""
		super().__init__()

		self.__file = None
		self.__key_modified = dict()

	def get_file(self):
		"""
		The method to get the file

		Returns:
			StreamWrapper: An instance of StreamWrapper
		"""

		return self.__file

	def set_file(self, file):
		"""
		The method to set the value to file

		Parameters:
			file (StreamWrapper) : An instance of StreamWrapper
		"""

		if file is not None and not isinstance(file, StreamWrapper):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: file EXPECTED TYPE: StreamWrapper', None, None)
		
		self.__file = file
		self.__key_modified['file'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification




try:
    import logging
    import os
    import json
    import threading
    from zcrmsdk.src.com.zoho.crm.api.user_signature import UserSignature
    from zcrmsdk.src.com.zoho.api.authenticator.store.token_store import TokenStore
    from zcrmsdk.src.com.zoho.crm.api.exception.sdk_exception import SDKException
    from zcrmsdk.src.com.zoho.crm.api.dc.data_center import DataCenter
    from zcrmsdk.src.com.zoho.crm.api.util.constants import Constants
    from zcrmsdk.src.com.zoho.api.authenticator.token import Token
    from zcrmsdk.src.com.zoho.api.logger import Logger, SDKLogger
    from zcrmsdk.src.com.zoho.crm.api.request_proxy import RequestProxy
    from zcrmsdk.src.com.zoho.crm.api.sdk_config import SDKConfig

except Exception:
    import logging
    import os
    import json
    import threading
    from ..api.user_signature import UserSignature
    from ...api.authenticator.store.token_store import TokenStore
    from ..api.exception import SDKException
    from ..api.dc.data_center import DataCenter
    from ..api.util.constants import Constants
    from ...api.authenticator.token import Token
    from ...api.logger import Logger, SDKLogger
    from .request_proxy import RequestProxy
    from .sdk_config import SDKConfig


class Initializer(object):

    """
    The class to initialize Zoho CRM SDK.
    """

    def __init__(self):
        self.environment = None
        self.user = None
        self.store = None
        self.token = None
        self.sdk_config = None
        self.request_proxy = None
        self.resource_path = None

    json_details = None
    initializer = None
    LOCAL = threading.local()
    LOCAL.init = None

    @staticmethod
    def initialize(user, environment, token, store, sdk_config, resource_path, logger=None, proxy=None):

        """
        The method to initialize the SDK.

        Parameters:
            user (UserSignature) : A UserSignature class instance represents the CRM user
            environment (DataCenter.Environment) : An Environment class instance containing the CRM API base URL and Accounts URL.
            token (Token) : A Token class instance containing the OAuth client application information.
            store (TokenStore) : A TokenStore class instance containing the token store information.
            sdk_config (SDKConfig) : A SDKConfig class instance containing the configuration.
            resource_path (str) : A String containing the absolute directory path to store user specific JSON files containing module fields information.
            logger (Logger): A Logger class instance containing the log file path and Logger type.
            proxy (RequestProxy) : A RequestProxy class instance containing the proxy properties of the user.
        """

        try:
            if not isinstance(user, UserSignature):
                error = {Constants.FIELD: Constants.USER, Constants.EXPECTED_TYPE: UserSignature.__module__}

                raise SDKException(Constants.INITIALIZATION_ERROR, Constants.INITIALIZATION_EXCEPTION, details=error)

            if not isinstance(environment, DataCenter.Environment):
                error = {Constants.FIELD: Constants.ENVIRONMENT,
                         Constants.EXPECTED_TYPE: DataCenter.Environment.__module__}

                raise SDKException(Constants.INITIALIZATION_ERROR, Constants.INITIALIZATION_EXCEPTION, details=error)

            if not isinstance(token, Token):
                error = {Constants.FIELD: Constants.TOKEN, Constants.EXPECTED_TYPE: Token.__module__}

                raise SDKException(Constants.INITIALIZATION_ERROR, Constants.INITIALIZATION_EXCEPTION, details=error)

            if not isinstance(store, TokenStore):
                error = {Constants.FIELD: Constants.STORE, Constants.EXPECTED_TYPE: TokenStore.__module__}

                raise SDKException(Constants.INITIALIZATION_ERROR, Constants.INITIALIZATION_EXCEPTION, details=error)

            if not isinstance(sdk_config, SDKConfig):
                error = {Constants.FIELD: Constants.SDK_CONFIG, Constants.EXPECTED_TYPE: SDKConfig.__module__}

                raise SDKException(Constants.INITIALIZATION_ERROR, Constants.INITIALIZATION_EXCEPTION, details=error)

            if resource_path is None or len(resource_path) == 0:
                raise SDKException(Constants.INITIALIZATION_ERROR, Constants.RESOURCE_PATH_ERROR_MESSAGE)

            if proxy is not None and not isinstance(proxy, RequestProxy):
                error = {Constants.FIELD: Constants.USER_PROXY, Constants.EXPECTED_TYPE: RequestProxy.__module__}

                raise SDKException(Constants.INITIALIZATION_ERROR, Constants.INITIALIZATION_EXCEPTION, details=error)

            if logger is not None:
                SDKLogger.initialize(logger)
            else:
                SDKLogger.initialize(Logger(Logger.Levels.INFO, os.path.join(os.getcwd(), Constants.LOGFILE_NAME)))

            try:
                json_details_path = os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', Constants.JSON_DETAILS_FILE_PATH)

                with open(json_details_path, mode='r') as JSON:
                    Initializer.json_details = json.load(JSON)
            except Exception as e:
                raise SDKException(code=Constants.JSON_DETAILS_ERROR, cause=e)

            initializer = Initializer()

            initializer.environment = environment
            initializer.user = user
            initializer.token = token
            initializer.store = store
            initializer.sdk_config = sdk_config
            initializer.resource_path = resource_path
            initializer.request_proxy = proxy

            Initializer.initializer = initializer

            logging.getLogger('SDKLogger').info(Constants.INITIALIZATION_SUCCESSFUL + initializer.__str__())

        except SDKException as e:
            raise e

    def __str__(self):
        return Constants.FOR_EMAIL_ID + Initializer.get_initializer().user.email + Constants.IN_ENVIRONMENT + Initializer.get_initializer().environment.url + '.'

    @staticmethod
    def get_initializer():

        """
        The method to get Initializer class instance.

        Returns:
            Initializer : An instance of Initializer
        """

        if getattr(Initializer.LOCAL, 'init', None) is not None:
            return getattr(Initializer.LOCAL, 'init')

        return Initializer.initializer

    @staticmethod
    def get_json(file_path):
        with open(file_path, mode="r") as JSON:
            file_contents = json.load(JSON)
            JSON.close()

        return file_contents

    @staticmethod
    def switch_user(user, environment, token, sdk_config, proxy=None):

        """
        The method to switch the different user in SDK environment.

        Parameters:
            user (UserSignature) : A UserSignature class instance represents the CRM user
            environment (DataCenter.Environment) : An Environment class instance containing the CRM API base URL and Accounts URL.
            token (Token) : A Token class instance containing the OAuth client application information.
            sdk_config (SDKConfig) : A SDKConfig class instance containing the configuration.
            proxy (RequestProxy) : A RequestProxy class instance containing the proxy properties of the user.
        """

        if not isinstance(user, UserSignature):
            error = {Constants.FIELD: Constants.USER, Constants.EXPECTED_TYPE: UserSignature.__module__}

            raise SDKException(Constants.SWITCH_USER_ERROR, Constants.SWITCH_USER_EXCEPTION, details=error)

        if not isinstance(environment, DataCenter.Environment):
            error = {Constants.FIELD: Constants.ENVIRONMENT,
                     Constants.EXPECTED_TYPE: DataCenter.Environment.__module__}

            raise SDKException(Constants.SWITCH_USER_ERROR, Constants.SWITCH_USER_EXCEPTION, details=error)

        if not isinstance(token, Token):
            error = {Constants.FIELD: Constants.TOKEN, Constants.EXPECTED_TYPE: Token.__module__}

            raise SDKException(Constants.SWITCH_USER_ERROR, Constants.SWITCH_USER_EXCEPTION, details=error)

        if not isinstance(sdk_config, SDKConfig):
            error = {Constants.FIELD: Constants.SDK_CONFIG, Constants.EXPECTED_TYPE: SDKConfig.__module__}

            raise SDKException(Constants.SWITCH_USER_ERROR, Constants.SWITCH_USER_EXCEPTION, details=error)

        if proxy is not None and not isinstance(proxy, RequestProxy):
            error = {Constants.FIELD: Constants.USER_PROXY, Constants.EXPECTED_TYPE: RequestProxy.__module__}

            raise SDKException(Constants.SWITCH_USER_ERROR, Constants.SWITCH_USER_EXCEPTION, details=error)

        initializer = Initializer()

        initializer.user = user
        initializer.environment = environment
        initializer.token = token
        initializer.sdk_config = sdk_config
        initializer.store = Initializer.initializer.store
        initializer.resource_path = Initializer.initializer.resource_path
        initializer.request_proxy = proxy

        Initializer.LOCAL.init = initializer

        logging.getLogger('SDKLogger').info(Constants.INITIALIZATION_SWITCHED + initializer.__str__())



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
	from zcrmsdk.src.com.zoho.crm.api.record.record import Record
except Exception:
	from ..exception import SDKException
	from ..util import Constants
	from .record import Record


class PricingDetails(Record):
	def __init__(self):
		"""Creates an instance of PricingDetails"""
		super().__init__()


	def get_to_range(self):
		"""
		The method to get the to_range

		Returns:
			float: A float representing the to_range
		"""

		return self.get_key_value('to_range')

	def set_to_range(self, to_range):
		"""
		The method to set the value to to_range

		Parameters:
			to_range (float) : A float representing the to_range
		"""

		if to_range is not None and not isinstance(to_range, float):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: to_range EXPECTED TYPE: float', None, None)
		
		self.add_key_value('to_range', to_range)

	def get_discount(self):
		"""
		The method to get the discount

		Returns:
			float: A float representing the discount
		"""

		return self.get_key_value('discount')

	def set_discount(self, discount):
		"""
		The method to set the value to discount

		Parameters:
			discount (float) : A float representing the discount
		"""

		if discount is not None and not isinstance(discount, float):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: discount EXPECTED TYPE: float', None, None)
		
		self.add_key_value('discount', discount)

	def get_from_range(self):
		"""
		The method to get the from_range

		Returns:
			float: A float representing the from_range
		"""

		return self.get_key_value('from_range')

	def set_from_range(self, from_range):
		"""
		The method to set the value to from_range

		Parameters:
			from_range (float) : A float representing the from_range
		"""

		if from_range is not None and not isinstance(from_range, float):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: from_range EXPECTED TYPE: float', None, None)
		
		self.add_key_value('from_range', from_range)



from .action_wrapper import ActionWrapper
from .related_list_properties import RelatedListProperties
from .argument import Argument
from .module import Module
from .body_wrapper import BodyWrapper
from .modules_operations import ModulesOperations, GetModulesHeader
from .api_exception import APIException
from .action_response import ActionResponse
from .response_handler import ResponseHandler
from .success_response import SuccessResponse
from .response_wrapper import ResponseWrapper
from .action_handler import ActionHandler
from .territory import Territory



from .api_exception import APIException
from .response_handler import ResponseHandler
from .territories_operations import TerritoriesOperations
from .response_wrapper import ResponseWrapper
from .territory import Territory



from abc import abstractmethod,  ABC


class DataCenter(ABC):

    """
    This class represents the properties of Zoho CRM DataCenter
    """

    @abstractmethod
    def get_iam_url(self):

        """
        The method to get the accounts URL.
        :return: A str representing the accounts URL.
        """

        pass

    @abstractmethod
    def get_file_upload_url(self):
        """
        The method to get the File Upload URL
        :return: A str representing the File Upload URL
        """

        pass

    class Environment(object):

        def __init__(self, url, accounts_url, file_upload_url):

            """
            Creates an Environment class instance with the specified parameters.
            :param url: A str representing the Zoho CRM API URL.
            :param accounts_url: A str representing the accounts URL.
            :param file_upload_url : A str representing the File Upload URL
            """

            self.url = url
            self.accounts_url = accounts_url
            self.file_upload_url = file_upload_url
            return



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Choice, Constants
except Exception:
	from ..exception import SDKException
	from ..util import Choice, Constants


class Criteria(object):
	def __init__(self):
		"""Creates an instance of Criteria"""

		self.__api_name = None
		self.__value = None
		self.__group_operator = None
		self.__group = None
		self.__comparator = None
		self.__key_modified = dict()

	def get_api_name(self):
		"""
		The method to get the api_name

		Returns:
			string: A string representing the api_name
		"""

		return self.__api_name

	def set_api_name(self, api_name):
		"""
		The method to set the value to api_name

		Parameters:
			api_name (string) : A string representing the api_name
		"""

		if api_name is not None and not isinstance(api_name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: api_name EXPECTED TYPE: str', None, None)
		
		self.__api_name = api_name
		self.__key_modified['api_name'] = 1

	def get_value(self):
		"""
		The method to get the value

		Returns:
			Object: A Object representing the value
		"""

		return self.__value

	def set_value(self, value):
		"""
		The method to set the value to value

		Parameters:
			value (Object) : A Object representing the value
		"""

		self.__value = value
		self.__key_modified['value'] = 1

	def get_group_operator(self):
		"""
		The method to get the group_operator

		Returns:
			Choice: An instance of Choice
		"""

		return self.__group_operator

	def set_group_operator(self, group_operator):
		"""
		The method to set the value to group_operator

		Parameters:
			group_operator (Choice) : An instance of Choice
		"""

		if group_operator is not None and not isinstance(group_operator, Choice):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: group_operator EXPECTED TYPE: Choice', None, None)
		
		self.__group_operator = group_operator
		self.__key_modified['group_operator'] = 1

	def get_group(self):
		"""
		The method to get the group

		Returns:
			list: An instance of list
		"""

		return self.__group

	def set_group(self, group):
		"""
		The method to set the value to group

		Parameters:
			group (list) : An instance of list
		"""

		if group is not None and not isinstance(group, list):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: group EXPECTED TYPE: list', None, None)
		
		self.__group = group
		self.__key_modified['group'] = 1

	def get_comparator(self):
		"""
		The method to get the comparator

		Returns:
			Choice: An instance of Choice
		"""

		return self.__comparator

	def set_comparator(self, comparator):
		"""
		The method to set the value to comparator

		Parameters:
			comparator (Choice) : An instance of Choice
		"""

		if comparator is not None and not isinstance(comparator, Choice):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: comparator EXPECTED TYPE: Choice', None, None)
		
		self.__comparator = comparator
		self.__key_modified['comparator'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



class SDKConfig(object):
    """
    The class to configure the SDK.
    """

    def __init__(self, auto_refresh_fields=False, pick_list_validation=True):
        """
        Creates an instance of SDKConfig with the following parameters

        Parameters:
            auto_refresh_fields(bool): A bool representing auto_refresh_fields
            pick_list_validation(bool): A bool representing pick_list_validation
        """
        self.__auto_refresh_fields = auto_refresh_fields
        self.__pick_list_validation = pick_list_validation

    def get_auto_refresh_fields(self):
        """
        This is a getter method to get auto_refresh_fields.

        Returns:
            bool: A bool representing auto_refresh_fields
        """
        return self.__auto_refresh_fields

    def get_pick_list_validation(self):
        """
        This is a getter method to get pick_list_validation.

        Returns:
            bool: A bool representing pick_list_validation
        """
        return self.__pick_list_validation



from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='zcrmsdk',
    version='3.1.0',
    description='Zoho CRM SDK for Python developers',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/zoho/zohocrm-python-sdk',
    author='Zoho CRM API Team',
    author_email='support@zohocrm.com',
    scripts=[],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    install_requires=[
        'requests',
        'python-dateutil',
        'urllib3'
    ],
    keywords=['development', 'zoho', 'crm', 'api', 'zcrmsdk', 'sdk', 'zcrm'],
    packages=find_packages(),
    include_package_data=True
)



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
	from zcrmsdk.src.com.zoho.crm.api.currencies.action_handler import ActionHandler
except Exception:
	from ..exception import SDKException
	from ..util import Constants
	from .action_handler import ActionHandler


class ActionWrapper(ActionHandler):
	def __init__(self):
		"""Creates an instance of ActionWrapper"""
		super().__init__()

		self.__currencies = None
		self.__key_modified = dict()

	def get_currencies(self):
		"""
		The method to get the currencies

		Returns:
			list: An instance of list
		"""

		return self.__currencies

	def set_currencies(self, currencies):
		"""
		The method to set the value to currencies

		Parameters:
			currencies (list) : An instance of list
		"""

		if currencies is not None and not isinstance(currencies, list):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: currencies EXPECTED TYPE: list', None, None)
		
		self.__currencies = currencies
		self.__key_modified['currencies'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification




try:
    import os
    import csv
    from zcrmsdk.src.com.zoho.api.authenticator.store.token_store import TokenStore
    from zcrmsdk.src.com.zoho.api.authenticator.oauth_token import OAuthToken, TokenType
    from ....crm.api.util.constants import Constants
    from zcrmsdk.src.com.zoho.crm.api.exception.sdk_exception import SDKException

except Exception as e:
    import os
    import csv
    from .token_store import TokenStore
    from ..oauth_token import OAuthToken, TokenType
    from ....crm.api.util.constants import Constants
    from zcrmsdk.src.com.zoho.crm.api.exception.sdk_exception import SDKException


class FileStore(TokenStore):

    """
    The class to store user token details to the file.
    """

    def __init__(self, file_path):

        """
        Creates an FileStore class instance with the specified parameters.

        Parameters:
            file_path (str) : A string containing the absolute file path of the file to store tokens

        """

        self.file_path = file_path
        self.headers = [Constants.USER_MAIL, Constants.CLIENT_ID, Constants.REFRESH_TOKEN, Constants.ACCESS_TOKEN, Constants.GRANT_TOKEN, Constants.EXPIRY_TIME]

        if (os.path.exists(file_path) and os.stat(file_path).st_size == 0) or not os.path.exists(file_path):
            with open(self.file_path, mode='w') as token_file:
                csv_writer = csv.writer(token_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                csv_writer.writerow(self.headers)

    def get_token(self, user, token):
        try:
            if isinstance(token, OAuthToken):
                with open(self.file_path, mode='r') as f:
                    data = csv.reader(f, delimiter=',')
                    next(data, None)
                    for next_record in data:
                        if len(next_record) == 0:
                            continue
                        if self.check_token_exists(user.email, token, next_record):
                            token.access_token = next_record[3]
                            token.expires_in = next_record[5]
                            token.refresh_token = next_record[2]
                            return token

        except IOError as ex:
            raise SDKException(code=Constants.TOKEN_STORE, message=Constants.GET_TOKEN_FILE_ERROR, cause=ex)

        return None

    def save_token(self, user, token):
        if isinstance(token, OAuthToken):
            token.user_mail = user.email
            self.delete_token(token)

            try:
                with open(self.file_path, mode='a+') as f:
                    csv_writer = csv.writer(f, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    csv_writer.writerow([user.email, token.client_id, token.refresh_token, token.access_token, token.grant_token, token.expires_in])

            except IOError as ex:
                raise SDKException(code=Constants.TOKEN_STORE, message=Constants.SAVE_TOKEN_FILE_ERROR, cause=ex)

    def delete_token(self, token):
        lines = list()

        if isinstance(token, OAuthToken):
            try:
                with open(self.file_path, mode='r') as f:
                    data = csv.reader(f, delimiter=',')
                    for next_record in data:
                        if len(next_record) == 0:
                            continue
                        lines.append(next_record)
                        if self.check_token_exists(token.user_mail, token, next_record):
                            lines.remove(next_record)

                with open(self.file_path, mode='w') as f:
                    csv_writer = csv.writer(f, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    csv_writer.writerows(lines)

            except IOError as ex:
                raise SDKException(code=Constants.TOKEN_STORE, message=Constants.DELETE_TOKEN_FILE_ERROR, cause=ex)

    def get_tokens(self):
        tokens = []

        try:
            with open(self.file_path, mode='r') as f:
                data = csv.reader(f, delimiter=',')
                next(data, None)
                for next_record in data:
                    if len(next_record) == 0:
                        continue
                    token_type = TokenType.REFRESH if len(next_record[4]) == 0 else TokenType.GRANT
                    token_value = next_record[4] if token_type == TokenType.GRANT else next_record[2]
                    token = OAuthToken(next_record[1], None, token_value, token_type)
                    token.user_mail = next_record[0]
                    token.expires_in = next_record[5]
                    token.access_token = next_record[3]
                    tokens.append(token)

            return tokens
        except Exception as ex:
            raise SDKException(code=Constants.TOKEN_STORE, message=Constants.GET_TOKENS_FILE_ERROR, cause=ex)

    def delete_tokens(self):
        try:
            with open(self.file_path, mode='w') as token_file:
                csv_writer = csv.writer(token_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                csv_writer.writerow(self.headers)
        except Exception as ex:
            raise SDKException(code=Constants.TOKEN_STORE, message=Constants.DELETE_TOKENS_FILE_ERROR, cause=ex)
    
    def check_token_exists(self, email, token, row):
        if email is None:
            raise SDKException(Constants.USER_MAIL_NULL_ERROR, Constants.USER_MAIL_NULL_ERROR_MESSAGE)

        client_id = token.client_id
        grant_token = token.grant_token
        refresh_token = token.refresh_token
        token_check = grant_token == row[4] if grant_token is not None else refresh_token == row[2]

        if row[0] == email and row[1] == client_id and token_check:
            return True
        
        return False




try:
    import json
    import platform
    import urllib3
    import logging
    from zcrmsdk.src.com.zoho.crm.api.util.api_http_connector import APIHTTPConnector
    from zcrmsdk.src.com.zoho.crm.api.util.json_converter import JSONConverter
    from zcrmsdk.src.com.zoho.crm.api.util.xml_converter import XMLConverter
    from zcrmsdk.src.com.zoho.crm.api.util.form_data_converter import FormDataConverter
    from zcrmsdk.src.com.zoho.crm.api.util.downloader import Downloader
    from zcrmsdk.src.com.zoho.crm.api.util.constants import Constants
    from zcrmsdk.src.com.zoho.crm.api.util.api_response import APIResponse
    from zcrmsdk.src.com.zoho.crm.api.header_map import HeaderMap
    from zcrmsdk.src.com.zoho.crm.api.header import Header
    from zcrmsdk.src.com.zoho.crm.api.parameter_map import ParameterMap
    from zcrmsdk.src.com.zoho.crm.api.param import Param
    from zcrmsdk.src.com.zoho.crm.api.exception import SDKException

except Exception:
    import json
    import platform
    import urllib3
    import logging
    from .api_http_connector import APIHTTPConnector
    from .json_converter import JSONConverter
    from .constants import Constants
    from .api_response import APIResponse
    from ..header_map import HeaderMap
    from ..header import Header
    from ..parameter_map import ParameterMap
    from ..param import Param
    from ..exception import SDKException


class CommonAPIHandler(object):
    """
    This class to process the API request and its response.
    Construct the objects that are to be sent as parameters or request body with the API.
    The Request parameter, header and body objects are constructed here.
    Process the response JSON and converts it to relevant objects in the library.
    """
    logger = logging.getLogger('SDKLogger')

    def __init__(self):

        self.__api_path = None
        self.__header = HeaderMap()
        self.__param = ParameterMap()
        self.__request = None
        self.__http_method = None
        self.__module_api_name = None
        self.__content_type = None
        self.__category_method = None
        self.__mandatory_checker = None

    def set_api_path(self, api_path):
        """
        The method to set the API Path

        Parameters:
            api_path(str) : A string containing the API Path
        """

        self.__api_path = api_path

    def get_api_path(self):
        """
        The method to get the API Path

        Returns:
            string: A string representing the API Path
        """
        return self.__api_path

    def set_header(self, header):
        """
        The method to set the API request header map.

        Parameters:
            header(HeaderMap): A HeaderMap class instance containing the API request headers
        """

        if header is None:
            return

        if self.__header.request_headers is not None and self.__header.request_headers:
            self.__header.request_headers.update(header.request_headers)
        else:
            self.__header = header

    def set_param(self, param):
        """
        The method to set the API request parameter map.

        Parameters:
            param(ParameterMap) : A ParameterMap class instance containing the API request parameters
        """

        if param is None:
            return

        if self.__param.request_parameters is not None and self.__param.request_parameters:
            self.__param.request_parameters.update(param.request_parameters)
        else:
            self.__param = param

    def set_request(self, request):
        """
        The method to set the request instance.

        Parameters:
            request(object): An object containing the request body

        """
        self.__request = request

    def set_http_method(self, http_method):
        """
        The method to set the HTTP Method

        Parameters:
            http_method(str):  A string containing the HTTP method.
        """

        self.__http_method = http_method

    def get_http_method(self):
        """
        The method to get the HTTP Method

        Returns:
            string: A string representing the HTTP Method
        """

        return self.__http_method

    def set_module_api_name(self, module_api_name):
        """
        The method to set the Module API Name

        Parameters:
            module_api_name(str):  A string containing the Module API Name
        """

        self.__module_api_name = module_api_name

    def get_module_api_name(self):
        """
        The method to get the Module API Name

        Returns:
            string: A string representing the Module API Name
        """

        return self.__module_api_name

    def set_content_type(self, content_type):
        """
        The method to set the Content Type

        Parameters:
            content_type(str):  A string containing the Content Type
        """

        self.__content_type = content_type

    def set_category_method(self, category_method):
        """
        The method to set the Category Method

        Parameters:
            category_method(str):  A string containing the Category method.
        """

        self.__category_method = category_method

    def get_category_method(self):
        """
        The method to get the Category Method

        Returns:
            string: A string representing the Category method.
        """

        return self.__category_method

    def set_mandatory_checker(self, mandatory_checker):
        """
        The method to set the Mandatory Checker

        Parameters:
            mandatory_checker(bool): A boolean containing the Mandatory Checker.
        """

        self.__mandatory_checker = mandatory_checker

    def get_mandatory_checker(self):
        """
        The method to get the Mandatory Checker

        Returns:
            bool: A boolean representing the Mandatory Checker.
        """
        return self.__mandatory_checker

    def add_param(self, param_instance, param_value):

        """
        The method to add an API request parameter.

        Parameters:
            param_instance (Param) : A Param instance containing the API request parameter.
            param_value (object) : An object containing the API request parameter value.
        """

        if param_value is None:
            return

        if self.__param is None:
            self.__param = ParameterMap()

        self.__param.add(param_instance, param_value)

    def add_header(self, header_instance, header_value):

        """
        The method to add an API request header.

        Parameters:
            header_instance (Header) : A Header instance containing the API request header.
            header_value (object) : An object containing the API request header value.
        """

        if header_value is None:
            return

        if self.__header is None:
            self.__header = HeaderMap()

        self.__header.add(header_instance, header_value)

    def api_call(self, class_name, encode_type):

        """
        The method to construct API request and response details. To make the Zoho CRM API calls.

        Parameters:
            class_name(str): A str containing the method return type.
            encode_type(str): A str containing the expected API response content type.

        Returns:
            APIResponse: An instance of APIResponse representing the Zoho CRM API response instance

        Raises:
            SDKException
        """

        try:
            from zcrmsdk.src.com.zoho.crm.api.initializer import Initializer
        except Exception:
            from ..initializer import Initializer

        if Initializer.get_initializer() is None:
            raise SDKException(code=Constants.SDK_UNINITIALIZATION_ERROR,
                               message=Constants.SDK_UNINITIALIZATION_MESSAGE)

        connector = APIHTTPConnector()
        try:
            self.set_api_url(connector)
        except SDKException as e:
            CommonAPIHandler.logger.error(Constants.SET_API_URL_EXCEPTION + e.__str__())
            raise e
        except Exception as e:
            sdk_exception = SDKException(cause=e)
            CommonAPIHandler.logger.error(Constants.SET_API_URL_EXCEPTION + sdk_exception.__str__())
            raise sdk_exception

        connector.request_method = self.__http_method
        connector.content_type = self.__content_type

        if self.__header is not None and len(self.__header.request_headers) > 0:
            connector.headers = self.__header.request_headers

        if self.__param is not None and len(self.__param.request_parameters) > 0:
            connector.parameters = self.__param.request_parameters

        try:
            Initializer.get_initializer().token.authenticate(connector)
        except SDKException as e:
            CommonAPIHandler.logger.info(Constants.AUTHENTICATION_EXCEPTION + e.__str__())
            raise e
        except Exception as e:
            sdk_exception = SDKException(cause=e)
            CommonAPIHandler.logger.error(Constants.AUTHENTICATION_EXCEPTION + sdk_exception.__str__())
            raise sdk_exception

        convert_instance = None

        if self.__content_type is not None and self.__http_method in [Constants.REQUEST_METHOD_PATCH, Constants.REQUEST_METHOD_POST, Constants.REQUEST_METHOD_PUT]:
            try:
                convert_instance = self.get_converter_class_instance(self.__content_type.lower())
                request = convert_instance.form_request(self.__request, self.__request.__class__.__module__, None, None)
            except SDKException as e:
                CommonAPIHandler.logger.info(Constants.FORM_REQUEST_EXCEPTION + e.__str__())
                raise e
            except Exception as e:
                sdk_exception = SDKException(cause=e)
                CommonAPIHandler.logger.error(Constants.FORM_REQUEST_EXCEPTION + sdk_exception.__str__())
                raise sdk_exception

            connector.request_body = request

        try:
            connector.headers[
                Constants.ZOHO_SDK] = platform.system() + "/" + platform.release() + " python/" + platform.python_version() + ":" + Constants.SDK_VERSION
            response = connector.fire_request(convert_instance)
            return_object = None

            if Constants.CONTENT_TYPE in response.headers:
                content_type = response.headers[Constants.CONTENT_TYPE]

                if ";" in content_type:
                    content_type = content_type.rpartition(";")[0]

                convert_instance = self.get_converter_class_instance(str(content_type).lower())
                return_object = convert_instance.get_wrapped_response(response, class_name)

            else:
                CommonAPIHandler.logger.info(response.__str__())

            return APIResponse(response.headers, response.status_code, return_object)
        except SDKException as e:
            CommonAPIHandler.logger.info(Constants.API_CALL_EXCEPTION + e.__str__())
        except Exception as e:
            sdk_exception = SDKException(cause=e)
            CommonAPIHandler.logger.error(Constants.API_CALL_EXCEPTION + sdk_exception.__str__())
            raise sdk_exception

    def get_converter_class_instance(self, encode_type):

        """
        This method to get a Converter class instance.
        :param encode_type: A str containing the API response content type.
        :return: A Converter class instance.
        """

        switcher = {

            "application/json": JSONConverter(self),

            "text/plain": JSONConverter(self),

            "application/ld+json": JSONConverter(self),

            "application/xml": XMLConverter(self),

            "text/xml": XMLConverter(self),

            "multipart/form-data": FormDataConverter(self),

            "application/x-download": Downloader(self),

            "image/png": Downloader(self),

            "image/jpeg": Downloader(self),

            "image/gif": Downloader(self),

            "image/tiff": Downloader(self),

            "image/svg+xml": Downloader(self),

            "image/bmp": Downloader(self),

            "image/webp": Downloader(self),

            "text/html": Downloader(self),

            "text/css": Downloader(self),

            "text/javascript": Downloader(self),

            "text/calendar": Downloader(self),

            "application/zip": Downloader(self),

            "application/pdf": Downloader(self),

            "application/java-archive": Downloader(self),

            "application/javascript": Downloader(self),

            "application/xhtml+xml": Downloader(self),

            "application/x-bzip": Downloader(self),

            "application/msword": Downloader(self),

            "application/vnd.openxmlformats-officedocument.wordprocessingml.document": Downloader(self),

            "application/gzip": Downloader(self),

            "application/x-httpd-php": Downloader(self),

            "application/vnd.ms-powerpoint": Downloader(self),

            "application/vnd.rar": Downloader(self),

            "application/x-sh": Downloader(self),

            "application/x-tar": Downloader(self),

            "application/vnd.ms-excel": Downloader(self),

            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": Downloader(self),

            "application/x-7z-compressed": Downloader(self),

            "audio/mpeg": Downloader(self),

            "audio/x-ms-wma": Downloader(self),

            "audio/vnd.rn-realaudio": Downloader(self),

            "audio/x-wav": Downloader(self),

            "audio/3gpp": Downloader(self),

            "audio/3gpp2": Downloader(self),

            "video/mpeg": Downloader(self),

            "video/mp4": Downloader(self),

            "video/webm": Downloader(self),

            "video/3gpp": Downloader(self),

            "video/3gpp2": Downloader(self),

            "font/ttf": Downloader(self),

            "text/csv": Downloader(self),

            "application/octet-stream": Downloader(self),
        }

        return switcher.get(encode_type, None)

    def set_api_url(self, connector):
        try:
            from zcrmsdk.src.com.zoho.crm.api.initializer import Initializer
        except Exception:
            from ..initializer import Initializer

        api_path = ''

        if Constants.HTTP in self.__api_path:
            if Constants.CONTENT_API_URL in self.__api_path:
                api_path = Initializer.get_initializer().environment.file_upload_url

                try:
                    url_parse = urllib3.util.parse_url(self.__api_path)
                    path = url_parse.path
                except Exception as ex:
                    raise SDKException(code=Constants.INVALID_URL_ERROR, cause=ex)

                api_path = api_path + path
            else:
                if str(self.__api_path)[:1].__eq__('/'):
                    self.__api_path = self.__api_path[1:]

                api_path = api_path + self.__api_path
        else:
            api_path = Initializer.get_initializer().environment.url
            api_path = api_path + self.__api_path

        connector.url = api_path



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class CustomizeInfo(object):
	def __init__(self):
		"""Creates an instance of CustomizeInfo"""

		self.__notes_desc = None
		self.__show_right_panel = None
		self.__bc_view = None
		self.__show_home = None
		self.__show_detail_view = None
		self.__unpin_recent_item = None
		self.__key_modified = dict()

	def get_notes_desc(self):
		"""
		The method to get the notes_desc

		Returns:
			bool: A bool representing the notes_desc
		"""

		return self.__notes_desc

	def set_notes_desc(self, notes_desc):
		"""
		The method to set the value to notes_desc

		Parameters:
			notes_desc (bool) : A bool representing the notes_desc
		"""

		if notes_desc is not None and not isinstance(notes_desc, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: notes_desc EXPECTED TYPE: bool', None, None)
		
		self.__notes_desc = notes_desc
		self.__key_modified['notes_desc'] = 1

	def get_show_right_panel(self):
		"""
		The method to get the show_right_panel

		Returns:
			string: A string representing the show_right_panel
		"""

		return self.__show_right_panel

	def set_show_right_panel(self, show_right_panel):
		"""
		The method to set the value to show_right_panel

		Parameters:
			show_right_panel (string) : A string representing the show_right_panel
		"""

		if show_right_panel is not None and not isinstance(show_right_panel, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: show_right_panel EXPECTED TYPE: str', None, None)
		
		self.__show_right_panel = show_right_panel
		self.__key_modified['show_right_panel'] = 1

	def get_bc_view(self):
		"""
		The method to get the bc_view

		Returns:
			string: A string representing the bc_view
		"""

		return self.__bc_view

	def set_bc_view(self, bc_view):
		"""
		The method to set the value to bc_view

		Parameters:
			bc_view (string) : A string representing the bc_view
		"""

		if bc_view is not None and not isinstance(bc_view, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: bc_view EXPECTED TYPE: str', None, None)
		
		self.__bc_view = bc_view
		self.__key_modified['bc_view'] = 1

	def get_show_home(self):
		"""
		The method to get the show_home

		Returns:
			bool: A bool representing the show_home
		"""

		return self.__show_home

	def set_show_home(self, show_home):
		"""
		The method to set the value to show_home

		Parameters:
			show_home (bool) : A bool representing the show_home
		"""

		if show_home is not None and not isinstance(show_home, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: show_home EXPECTED TYPE: bool', None, None)
		
		self.__show_home = show_home
		self.__key_modified['show_home'] = 1

	def get_show_detail_view(self):
		"""
		The method to get the show_detail_view

		Returns:
			bool: A bool representing the show_detail_view
		"""

		return self.__show_detail_view

	def set_show_detail_view(self, show_detail_view):
		"""
		The method to set the value to show_detail_view

		Parameters:
			show_detail_view (bool) : A bool representing the show_detail_view
		"""

		if show_detail_view is not None and not isinstance(show_detail_view, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: show_detail_view EXPECTED TYPE: bool', None, None)
		
		self.__show_detail_view = show_detail_view
		self.__key_modified['show_detail_view'] = 1

	def get_unpin_recent_item(self):
		"""
		The method to get the unpin_recent_item

		Returns:
			string: A string representing the unpin_recent_item
		"""

		return self.__unpin_recent_item

	def set_unpin_recent_item(self, unpin_recent_item):
		"""
		The method to set the value to unpin_recent_item

		Parameters:
			unpin_recent_item (string) : A string representing the unpin_recent_item
		"""

		if unpin_recent_item is not None and not isinstance(unpin_recent_item, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: unpin_recent_item EXPECTED TYPE: str', None, None)
		
		self.__unpin_recent_item = unpin_recent_item
		self.__key_modified['unpin_recent_item'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class Crypt(object):
	def __init__(self):
		"""Creates an instance of Crypt"""

		self.__mode = None
		self.__column = None
		self.__encfldids = None
		self.__notify = None
		self.__table = None
		self.__status = None
		self.__key_modified = dict()

	def get_mode(self):
		"""
		The method to get the mode

		Returns:
			string: A string representing the mode
		"""

		return self.__mode

	def set_mode(self, mode):
		"""
		The method to set the value to mode

		Parameters:
			mode (string) : A string representing the mode
		"""

		if mode is not None and not isinstance(mode, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: mode EXPECTED TYPE: str', None, None)
		
		self.__mode = mode
		self.__key_modified['mode'] = 1

	def get_column(self):
		"""
		The method to get the column

		Returns:
			string: A string representing the column
		"""

		return self.__column

	def set_column(self, column):
		"""
		The method to set the value to column

		Parameters:
			column (string) : A string representing the column
		"""

		if column is not None and not isinstance(column, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: column EXPECTED TYPE: str', None, None)
		
		self.__column = column
		self.__key_modified['column'] = 1

	def get_encfldids(self):
		"""
		The method to get the encfldids

		Returns:
			list: An instance of list
		"""

		return self.__encfldids

	def set_encfldids(self, encfldids):
		"""
		The method to set the value to encfldids

		Parameters:
			encfldids (list) : An instance of list
		"""

		if encfldids is not None and not isinstance(encfldids, list):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: encfldids EXPECTED TYPE: list', None, None)
		
		self.__encfldids = encfldids
		self.__key_modified['encFldIds'] = 1

	def get_notify(self):
		"""
		The method to get the notify

		Returns:
			string: A string representing the notify
		"""

		return self.__notify

	def set_notify(self, notify):
		"""
		The method to set the value to notify

		Parameters:
			notify (string) : A string representing the notify
		"""

		if notify is not None and not isinstance(notify, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: notify EXPECTED TYPE: str', None, None)
		
		self.__notify = notify
		self.__key_modified['notify'] = 1

	def get_table(self):
		"""
		The method to get the table

		Returns:
			string: A string representing the table
		"""

		return self.__table

	def set_table(self, table):
		"""
		The method to set the value to table

		Parameters:
			table (string) : A string representing the table
		"""

		if table is not None and not isinstance(table, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: table EXPECTED TYPE: str', None, None)
		
		self.__table = table
		self.__key_modified['table'] = 1

	def get_status(self):
		"""
		The method to get the status

		Returns:
			int: An int representing the status
		"""

		return self.__status

	def set_status(self, status):
		"""
		The method to set the value to status

		Parameters:
			status (int) : An int representing the status
		"""

		if status is not None and not isinstance(status, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: status EXPECTED TYPE: int', None, None)
		
		self.__status = status
		self.__key_modified['status'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import APIResponse, CommonAPIHandler, Constants
except Exception:
	from ..exception import SDKException
	from ..util import APIResponse, CommonAPIHandler, Constants


class BluePrintOperations(object):
	def __init__(self, record_id, module_api_name):
		"""
		Creates an instance of BluePrintOperations with the given parameters

		Parameters:
			record_id (int) : An int representing the record_id
			module_api_name (string) : A string representing the module_api_name
		"""

		if not isinstance(record_id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: record_id EXPECTED TYPE: int', None, None)
		
		if not isinstance(module_api_name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: module_api_name EXPECTED TYPE: str', None, None)
		
		self.__record_id = record_id
		self.__module_api_name = module_api_name


	def get_blueprint(self):
		"""
		The method to get blueprint

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2/'
		api_path = api_path + str(self.__module_api_name)
		api_path = api_path + '/'
		api_path = api_path + str(self.__record_id)
		api_path = api_path + '/actions/blueprint'
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_GET)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_READ)
		try:
			from zcrmsdk.src.com.zoho.crm.api.blue_print.response_handler import ResponseHandler
		except Exception:
			from .response_handler import ResponseHandler
		return handler_instance.api_call(ResponseHandler.__module__, 'application/json')

	def update_blueprint(self, request):
		"""
		The method to update blueprint

		Parameters:
			request (BodyWrapper) : An instance of BodyWrapper

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.blue_print.body_wrapper import BodyWrapper
		except Exception:
			from .body_wrapper import BodyWrapper

		if request is not None and not isinstance(request, BodyWrapper):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: request EXPECTED TYPE: BodyWrapper', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2/'
		api_path = api_path + str(self.__module_api_name)
		api_path = api_path + '/'
		api_path = api_path + str(self.__record_id)
		api_path = api_path + '/actions/blueprint'
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_PUT)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_UPDATE)
		handler_instance.set_content_type('application/json')
		handler_instance.set_request(request)
		handler_instance.set_mandatory_checker(True)
		try:
			from zcrmsdk.src.com.zoho.crm.api.blue_print.action_response import ActionResponse
		except Exception:
			from .action_response import ActionResponse
		return handler_instance.api_call(ActionResponse.__module__, 'application/json')
